package de.tu_darmstadt.tk.SmartHomeNetworkSim.core;

import java.util.LinkedList;
import java.util.Observable;

import javax.swing.Timer;

import de.tu_darmstadt.tk.SmartHomeNetworkSim.control.Controller;

/**
 * Manages the Simulation by calling the relevant methods.
 *
 * @author Andreas T. Meyer-Berg
 */
public class SimulationManager extends Observable {
	/**
	 * Model which should be simulated
	 */
	private Model model;

	/**
	 * Controller for manipulationAlgorithms
	 */
	private Controller controller;

	/**
	 * Whether the model status changed during the last simulation and the panel
	 * should be repainted
	 */
	private boolean statusChanged = false;
	
	/**
	 * Timer which triggers each simulation step
	 */
	private Timer timer;

	/**
	 * First Timestep of the simulation
	 */
	private long startTime = 0L;
	
	/**
	 * Last Timestep of the simulation
	 */
	private long endTime = 1000L;
	
	/**
	 * Current TimeStep in the simulation
	 */
	private long currentTime = 0L;
	
	/**
	 * Duration of the simulation
	 */
	private long duration = 100L;
	
	/**
	 * ManipulationAlgorithm Instances, which should run in each timestep
	 */
	private LinkedList<NetworkManipulationAlgorithm> algos = new LinkedList<NetworkManipulationAlgorithm>();
	
	/**
	 * PacketCollectionManager which manages the collection of packets
	 */
	private PacketCollectionManager collectionMan;
	
	/**
	 * PacketExportManager which handles packet export
	 */
	private PacketExportManager exportMan;
	
	/**
	 * Creates a new Simulationmanager
	 * 
	 * @param model
	 *            Model that should be simulated
	 */
	public SimulationManager(Model model) {
		this.model = model;
		this.controller = null;
		this.collectionMan = new PacketCollectionManager(model);
		this.exportMan = new PacketExportManager(model);
		timer = new Timer(0, t -> simulateTimeStep());
	}

	/**
	 * Simulate the next Timestep
	 */
	private void simulateTimeStep() {
		// Just simulate if timer is running
		if (endTime <= currentTime){
			stopSimulation();
			return;
		}
		if (currentTime + duration <= endTime) {
			simulateTimeIntervall(currentTime, duration);
			currentTime += duration;
			notifyPanels();
		} else {
			simulateTimeIntervall(currentTime, endTime - currentTime);
			currentTime = endTime;
			stopSimulation();
		}
	}

	/**
	 * Simulates one time step at a given time for a given duration
	 * 
	 * @param startTime
	 *            Time the simulation interval starts in
	 *            System.currentTimeMillis() time
	 * @param duration
	 *            Duration of the simulation interval in milliseconds
	 */
	public void simulateTimeIntervall(long startTime, long duration) {
		//Simulates the network
		simulateNetwork(startTime, duration);
		runAlgorithms(startTime+duration);
		collectionMan.collectPackets();
		//Run all 
		collectionMan.runPacketAlgorithms();
		//Export Packets
		exportMan.exportPacketsOfLastTimeStep();
	}
	
	/**
	 * Simulates one time step of the network at a given time for a given duration
	 * 
	 * @param startTime
	 *            Time the simulation interval starts in
	 *            System.currentTimeMillis() time
	 * @param duration
	 *            Duration of the simulation interval in milliseconds
	 */
	private void simulateNetwork(long startTime, long duration){
		// Nothing changed so far
		statusChanged = false;
		// Simulate all Links, and their connections
		model.getConnectionNetworks().forEach(d -> {
			d.simulateTimeInterval(startTime, duration);
			if (d.getStatusChanged()) {
				for (Connection c : d.getConnections()) {
					if (c.getStatus() == Connection.DONE) {
						model.getConnections().remove(c);
					}
				}
				statusChanged = true;
			}
		});
		// Simulate SmartDevices - if they need some logic
		model.getDevices().forEach(d -> d.simulateTimeStep(startTime, duration));
		// Store Packages/Export Packages etc. (for debug purposes)
		if (statusChanged) {
			model.setChanged();
			model.notifyObservers();
		}
	}

	/**
	 * Start the simulation
	 */
	public void startSimulation() {
		timer.start();
		notifyPanels();
	}

	/**
	 * Stop the simulation
	 */
	public void stopSimulation() {
		timer.stop();
		notifyPanels();
	}

	/**
	 * Resets the Simulation to the start time
	 */
	public void resetSimulation() {
		timer.stop();
		timer = new Timer(0, a -> simulateTimeStep());
		currentTime = startTime;
		resetSimulation(currentTime);
		notifyPanels();
	}

	/**
	 * Reset Simulation
	 * 
	 * @param timestep new timestep of ports after the reset
	 */
	private void resetSimulation(long timestep) {
		for (SmartDevice d : model.getDevices())
			for (Port p : d.getPorts())
				if (p.getLastTrigger() > timestep)
					p.setLastTrigger(timestep);
	}

	/**
	 * Returns true if the simulation is running, false if not
	 * 
	 * @return true if running
	 */
	public boolean isRunning() {
		return timer.isRunning();
	}

	/**
	 * Returns the StartTime of the simulation
	 * 
	 * @return startTime
	 */
	public long getStartTime() {
		return startTime;
	}

	/**
	 * Sets the new startTime
	 * 
	 * @param startTime
	 *            time the simulations starts
	 */
	public void setStartTime(long startTime) {
		this.startTime = startTime;
		notifyPanels();
	}

	/**
	 * Returns the end time of the simulation
	 * 
	 * @return End time of the simulation
	 */
	public long getEndTime() {
		return endTime;
	}

	/**
	 * Sets the new startTime
	 * @param endTime new EndTime
	 */
	public void setEndTime(long endTime) {
		this.endTime = endTime;
		notifyPanels();
	}

	/**
	 * @return the currentTime
	 */
	public long getCurrentTime() {
		return currentTime;
	}

	/**
	 * @param currentTime
	 *            the currentTime to set
	 */
	public void setCurrentTime(long currentTime) {
		this.currentTime = currentTime;
		notifyPanels();
	}

	/**
	 * Returns the simulation step duration in milliseconds
	 * 
	 * @return duration of each simulation step in milliseconds
	 */
	public long getStepDuration() {
		return duration;
	}

	/**
	 * Sets the duration of simulation steps
	 * 
	 * @param duration
	 *            duration in milliseconds of a step
	 */
	public void setStepDuration(long duration) {
		if(duration>0)
			this.duration = duration;
		else
			this.duration = 1;
		notifyPanels();
	}

	/**
	 * Notify the panels, which could update their GUI
	 */
	public void notifyPanels() {
		this.setChanged();
		this.notifyObservers();
	}
	
	/**
	 * Runs all registered algorithms at the currentTimeStep
	 * @param time currentTime of the simulation
	 */
	public void runAlgorithms(long time) {
		/**
		 * Run all Algorithms
		 */
		for(NetworkManipulationAlgorithm algo:algos){
			algo.runAlgorithm(controller, time);
		}
	}
	/**
	 * Returns all registered algorithms of the simulation
	 * @return all registered algorithms
	 */
	public LinkedList<NetworkManipulationAlgorithm> getAlgorithms(){
		return algos;
	}
	/**
	 * Adds an algorithm, which should be executed each timestep
	 * @param algo new algorithm
	 */
	public void addAlgorithm(NetworkManipulationAlgorithm algo, Controller controller){
		this.controller = controller;
		if(algo!=null)
			algos.add(algo);
	}
	
	/**
	 * Removes algorithms from the simulation
	 * @param algo algorithm to be removed
	 */
	public void removeAlgo(NetworkManipulationAlgorithm algo){
		algos.remove(algo);
	}
	
	/**
	 * Returns the PacketCollectionManager of the simulation
	 * @return PacketCollectionManager, which contains all the different packet collectors
	 */
	public PacketCollectionManager getPacketCollectionManager(){
		return collectionMan;
	}
	
	/**
	 * Returns the PacketExportManager
	 * @return the exportMan
	 */
	public PacketExportManager getPacketExportManager(){
		return exportMan;
	}
}
